/*
 * Name:    mp_field.c
 *
 * Purpose: MX Python private interface to MX record fields.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 2002-2003 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "Python.h"

#include "mx_util.h"
#include "mx_record.h"
#include "mx_handle.h"
#include "mp.h"

typedef struct {
	PyObject **object_array;
} MPPRIVATE_TUPLE_DATA_STRUCT;

static mx_status_type create_field_from_tuple_handler( MX_RECORD *record,
						MX_RECORD_FIELD *record_field,
						void *handler_data_ptr,
						long *array_indices,
						void *array_ptr,
						long dimension_level );

static mx_status_type create_tuple_from_field_handler( MX_RECORD *record,
						MX_RECORD_FIELD *record_field,
						void *handler_data_ptr,
						long *array_indices,
						void *array_ptr,
						long dimension_level );

/*---*/

MX_EXPORT mx_status_type
MpPrivate_create_field_from_tuple( MX_RECORD *record,
				MX_RECORD_FIELD *field,
				PyObject *tuple )
{
	const char fname[] = "MpPrivate_create_field_from_tuple()";

	MPPRIVATE_TUPLE_DATA_STRUCT tuple_data;
	long *array_indices;
	long i, n, object_array_length;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}
	if ( field == (MX_RECORD_FIELD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD_FIELD pointer passed was NULL." );
	}

	MX_DEBUG( 2,("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"));

	MX_DEBUG( 2,("%s invoked for record '%s', field '%s'.",
			fname, record->name, field->name ));

	MX_DEBUG( 2,("%s: field->datatype = %s, field->num_dimensions = %ld",
			fname, mx_get_field_type_string( field->datatype ),
			field->num_dimensions ));

	for ( i = 0; i < field->num_dimensions; i++ ) {

		n = field->num_dimensions - i - 1;

		MX_DEBUG( 2,
	("%s: field->dimension[%ld] = %ld, data_element_size[%ld] = %ld",
			fname, i, field->dimension[i],
			n, (long) field->data_element_size[n] ));
	}

	if ( PyTuple_Check( tuple ) == 0 ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The object %p passed for record '%s', field '%s' is not a tuple.",
			tuple, record->name, field->name );
	}

	/* Allocate a 1-D array to hold the field_array_indices. */

	array_indices = (long *) malloc( field->num_dimensions * sizeof(long) );

	if ( array_indices == (long *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %ld element "
		"array of field array pointers.", field->num_dimensions );
	}

	/* Allocate an array to contain pointers to the current object
	 * for each level in the array.  For example, if 
	 * create_field_from_tuple_handler() is invoked on array element
	 * [2][4][5], then object_array[0] will stand for the top level
	 * row of row pointers, object_array[0] will contain the top level
	 * tuple for the data as a whole, object_array[1] will contain the
	 * row tuple for row[2], object_array[2] will contain the row tuple
	 * for row[2][4], object_array[3] will contain the row tuple for
	 * row[2][4][5], and so forth.
	 *
	 * Note that object_array_length is incremented by 1 to add a slot
	 * for the pointer to the Python object that represents the field
	 * as a whole.
	 */

	object_array_length = field->num_dimensions + 1;

	tuple_data.object_array = (PyObject **)
			malloc( object_array_length * sizeof(PyObject *) );

	if ( tuple_data.object_array == (PyObject **) NULL ) {
		mx_free( array_indices );

		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %ld element "
		"array of tuple pointers.", field->num_dimensions );
	}

	/* Initialize the first element of the array to the Python
	 * tuple that we were passed.
	 */

	tuple_data.object_array[0] = tuple;

	/* Initialize the rest of the array to all NULL pointers. */

	for ( i = 1; i < object_array_length; i++ ) {
		tuple_data.object_array[i] = NULL;
	}

	MX_DEBUG( 2,("%s: tuple_data.object_array = %p",
		fname, tuple_data.object_array));

	MX_DEBUG( 2,("%s: tuple_data.object_array[0] = %p",
		fname, tuple_data.object_array[0]));

	/* Walk through all the elements of the tuple. */

	mx_status = mx_traverse_field( record, field,
					create_field_from_tuple_handler,
					&tuple_data,
					array_indices );

	mx_free( array_indices );

	return mx_status;
}

MX_EXPORT mx_status_type
MpPrivate_create_tuple_from_field( MX_RECORD *record,
				MX_RECORD_FIELD *field,
				PyObject **new_tuple )
{
	const char fname[] = "MpPrivate_create_tuple_from_field()";

	MPPRIVATE_TUPLE_DATA_STRUCT tuple_data;
	char field_description_buffer[1000];
	long *array_indices;
	long i, n, object_array_length;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		Py_INCREF(Py_None);
		*new_tuple = Py_None;

		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}
	if ( field == (MX_RECORD_FIELD *) NULL ) {
		Py_INCREF(Py_None);
		*new_tuple = Py_None;

		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD_FIELD pointer passed was NULL." );
	}

	MX_DEBUG( 2,("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"));

	MX_DEBUG( 2,("%s invoked for record '%s', field '%s'.",
			fname, record->name, field->name ));

	MX_DEBUG( 2,("%s: field->datatype = %s, field->num_dimensions = %ld",
			fname, mx_get_field_type_string( field->datatype ),
			field->num_dimensions ));

	for ( i = 0; i < field->num_dimensions; i++ ) {

		n = field->num_dimensions - i - 1;

		MX_DEBUG( 2,
	("%s: field->dimension[%ld] = %ld, data_element_size[%ld] = %ld",
			fname, i, field->dimension[i],
			n, (long) field->data_element_size[n] ));
	}

	MX_DEBUG( 2,("%s: new_tuple = %p", fname, new_tuple));

	MX_DEBUG( 2,("%s: *new_tuple = %p", fname, *new_tuple));

#if 1
	MX_DEBUG( 2,("%s: About to invoke mx_create_description_from_field()",
		fname));

	mx_status = mx_create_description_from_field( record, field,
			field_description_buffer,
			sizeof( field_description_buffer ) - 1 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s: field description = '%s'",
		fname, field_description_buffer ));
#endif

	/* Allocate a 1-D array to hold the field_array_indices. */

	array_indices = (long *) malloc( field->num_dimensions * sizeof(long) );

	if ( array_indices == (long *) NULL ) {
		Py_INCREF(Py_None);
		*new_tuple = Py_None;

		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %ld element "
		"array of field array pointers.", field->num_dimensions );
	}

	/* Allocate an array to contain pointers to the current tuple
	 * for each level in the array.  For example, if 
	 * create_tuple_from_field_handler() is invoked on array element
	 * [2][4][5], then object_array[0] will stand for the top level
	 * row of row pointers, object_array[0] will contain the top level
	 * tuple for the data as a whole, object_array[1] will contain the
	 * row tuple for row[2], object_array[2] will contain the row tuple
	 * for row[2][4], object_array[3] will contain the row tuple for
	 * row[2][4][5], and so forth.
	 *
	 * Note that object_array_length is incremented by 1 to add a slot
	 * for the pointer to the Python object that represents the field
	 * as a whole.
	 */

	object_array_length = field->num_dimensions + 1;

	tuple_data.object_array = (PyObject **)
			malloc( object_array_length * sizeof(PyObject *) );

	if ( tuple_data.object_array == (PyObject **) NULL ) {
		mx_free( array_indices );

		Py_INCREF(Py_None);
		*new_tuple = Py_None;

		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %ld element "
		"array of tuple pointers.", field->num_dimensions );
	}

	/* Initialize the array to all NULL pointers. */

	for ( i = 0; i < object_array_length; i++ ) {
		tuple_data.object_array[i] = NULL;
	}

	/* Now walk through all the elements of the field data array. */

	mx_status = mx_traverse_field( record, field,
					create_tuple_from_field_handler,
					&tuple_data,
					array_indices );

	MX_DEBUG( 2,("%s: returned from mx_traverse_field()", fname));

	MX_DEBUG( 2,("%s: tuple_data.object_array = %p",
			fname, tuple_data.object_array));

	MX_DEBUG( 2,("%s: tuple_data.object_array[0] = %p",
			fname, tuple_data.object_array[0]));

	if ( mx_status.code != MXE_SUCCESS ) {
		Py_INCREF(Py_None);
		*new_tuple = Py_None;
	} else {
		*new_tuple = tuple_data.object_array[0];
	}

	MX_DEBUG( 2,("About to free array_indices"));

	mx_free( array_indices );

	MX_DEBUG( 2,("About to free tuple_data.object_array = %p",
			tuple_data.object_array));

#if 0
	mx_free( tuple_data.object_array );
#endif

	MX_DEBUG( 2,("%s: successfully freed tuple_data.object_array", fname));

	MX_DEBUG( 2,("%s: new_tuple = %p", fname, new_tuple));

	MX_DEBUG( 2,("%s: returning Python object %p", fname, *new_tuple));

	return mx_status;
}

/*---*/

static mx_status_type
create_field_from_tuple_handler( MX_RECORD *record,
				MX_RECORD_FIELD *field,
				void *handler_data_ptr,
				long *array_indices,
				void *array_ptr,
				long dimension_level )
{
	const char fname[] = "create_field_from_tuple_handler()";

	MPPRIVATE_TUPLE_DATA_STRUCT *tuple_data;
	PyObject **object_array;
	PyObject *parent_tuple, *current_object;
	int at_bottom_level;
	long i, k, m, num_elements;
	char *pystring_ptr;
	char *string_ptr;
	char *char_ptr;
	unsigned char *uchar_ptr;
	short *short_ptr;
	unsigned short *ushort_ptr;
	int *int_ptr;
	unsigned int *uint_ptr;
	long *long_ptr;
	unsigned long *ulong_ptr;
	unsigned long *hex_ptr;
	float *float_ptr;
	double *double_ptr;

	MX_DEBUG( 2,("++++++ create_field_from_tuple_handler ++++++"));

	MX_DEBUG( 2,("%s: record = '%s', field = '%s', DIMENSION_LEVEL = %ld",
		fname, record->name, field->name, dimension_level ));

	if ( handler_data_ptr == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"handler_data_ptr is a NULL pointer." );
	}
	if ( array_indices == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"array_indices is a NULL pointer." );
	}
	if ( array_ptr == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"array_ptr is a NULL pointer." );
	}

	MX_DEBUG( 2,("%s: handler_data_ptr = %p", fname, handler_data_ptr));

	tuple_data = (MPPRIVATE_TUPLE_DATA_STRUCT *) handler_data_ptr;

	MX_DEBUG( 2,("%s: tuple_data = %p", fname, tuple_data));

	object_array = tuple_data->object_array;

	MX_DEBUG( 2,("%s: object_array = %p", fname, object_array));

	MX_DEBUG( 2,("%s: object_array[0] = %p", fname, object_array[0]));

#if 1
	if ( field->num_dimensions < 2 ) {
		MX_DEBUG( 2,("%s: field->num_dimensions < 2", fname));
	} else {
		MX_DEBUG( 2,
		("%s: object_array[1] = %p", fname, object_array[1]));
	}
#endif

	/*
	 * Guide to array index use:
	 *
	 * m = field->num_dimensions - dimension_level
	 *
	 * object_array[m] = This is the object that we are currently using.
	 *
	 * object_array[m-1] = This is the parent tuple of the object that
	 *                    we are currently using unless m == 0 (which
	 *                    means we are at the top level object).
	 *
	 * k = array_indices[m-1] = This tells us what index value to use 
	 *                      when finding this object in the parent tuple
	 *                      as in
	 *
	 *       PyTuple_GetItem( object_array[m-1], k )
	 * 
	 * num_elements = field->dimension[m] = Number of elements in the
         *                                      current dimension.
	 *
	 */

	m = field->num_dimensions - dimension_level;

	if ( m > 0 ) {
		parent_tuple = object_array[m-1];

		if ( PyTuple_Check(parent_tuple) == 0 ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
				"The parent_tuple object %p for m = %ld, "
				"dimension_level = %ld is not a tuple.  "
				"This should never happen and is a "
				"program bug.", parent_tuple, m,
				dimension_level );
		}

		k = array_indices[m-1];
	} else
	if ( m == 0 ) {
		parent_tuple = NULL;

		k = -1;
	} else {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"%s was invoked at dimension level %ld for record '%s', "
		"field '%s' which is larger than the total number of "
		"dimensions %ld.  This should _never_ happen and should "
		"be reported as a bug.", fname, dimension_level,
		record->name, field->name, field->num_dimensions );
	}

	MX_DEBUG( 2,("%s: m = %ld, parent_tuple = %p", fname, m, parent_tuple));

	MX_DEBUG( 2,("%s: field->num_dimensions = %ld, m = %ld, k = %ld",
			fname, field->num_dimensions, m, k));

	for ( i = 0; i < field->num_dimensions; i++ ) {
		MX_DEBUG( 2,(
"%s: array_indices[%ld] = %ld, dimension[%ld] = %ld, object_array[%ld] = %p",
			fname, i, array_indices[i], i, field->dimension[i],
			i, object_array[i]));
	}

	/* Get the current object at this level. */

	if ( m == 0 ) {
		current_object = object_array[0];
	} else {
		current_object = PyTuple_GetItem( parent_tuple, k );

		object_array[m] = current_object;

		if ( current_object == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The function PyTuple_GetItem( %p, %ld ) returned a NULL pointer.  "
	"This should never happen and is a program bug.",
				parent_tuple, m );
		}
	}

	MX_DEBUG( 2,("%s: current_object = %p", fname, current_object));

	/* Figure out whether or not we have reached the bottom of 
	 * the field data structure or not.
	 */

	at_bottom_level = FALSE;

	if ( dimension_level < 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"%s was invoked at dimension level %ld < 0 for record '%s', "
		"field '%s'.  This should _never_ happen and should be "
		"reported as a bug.", fname, dimension_level,
		record->name, field->name );

	} else
	if ( dimension_level > 1 ) {
		at_bottom_level = FALSE;

	} else
	if ( dimension_level == 1 ) {
		if ( field->datatype == MXFT_STRING ) {
			at_bottom_level = TRUE;
		} else {
			at_bottom_level = FALSE;
		}
	} else
	if ( dimension_level == 0 ) {
		if ( field->datatype == MXFT_STRING ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"%s was invoked at dimension level 0 for record '%s' "
			"and string field '%s'.  This should _never_ happen "
			"and should be reported as a bug.", fname,
			record->name, field->name );
		}
		at_bottom_level = TRUE;
	}

	MX_DEBUG( 2,("%s: at_bottom_level = %d", fname, at_bottom_level));

	if ( at_bottom_level == FALSE ) {

		/* If we are not at the bottom level yet, saving current_object
		 * to object_array[m] as we did earlier is all we need
		 * to do for this invocation.
		 */

	} else {
		/* at_bottom_level == TRUE */

		MX_DEBUG( 2,("%s: field->datatype = %ld, array_ptr = %p",
			fname, field->datatype, array_ptr));

		/* We have reached the bottom level and need to copy
		 * the value in the current object into the record field.
		 */
		switch ( field->datatype ) {
		case MXFT_STRING:
			string_ptr = (char *) array_ptr;

			MX_DEBUG( 2,("%s: STRING PTR = %p",
					fname, string_ptr));

			num_elements = field->dimension[m];

			MX_DEBUG( 2,("%s: num_elements = %ld",
					fname, num_elements));

			pystring_ptr = PyString_AsString( current_object );

			MX_DEBUG( 2,("%s: pystring_ptr = %p",
					fname, pystring_ptr ));

			MX_DEBUG( 2,("%s: pystring_ptr = '%s'",
					fname, pystring_ptr ));

			mx_strncpy( string_ptr, pystring_ptr, num_elements );

			MX_DEBUG( 2,("%s: string value = '%s'",
					fname, string_ptr));

			break;
		case MXFT_CHAR:
			char_ptr = (char *) array_ptr;

			*char_ptr = (char) PyInt_AsLong( current_object );

			MX_DEBUG( 2,("%s: CHAR VALUE = %d",
					fname, (int) *char_ptr));
			break;
		case MXFT_UCHAR:
			uchar_ptr = (unsigned char *) array_ptr;

			*uchar_ptr = (unsigned char)
					PyInt_AsLong( current_object );

			MX_DEBUG( 2,("%s: UCHAR VALUE = %d",
					fname, (int) *uchar_ptr));
			break;
		case MXFT_SHORT:
			short_ptr = (short *) array_ptr;

			*short_ptr = (short) PyInt_AsLong( current_object );

			MX_DEBUG( 2,("%s: SHORT VALUE = %hd",
					fname, *short_ptr));
			break;
		case MXFT_USHORT:
			ushort_ptr = (unsigned short *) array_ptr;

			*ushort_ptr = (unsigned short)
					PyInt_AsLong( current_object );

			MX_DEBUG( 2,("%s: USHORT VALUE = %hu",
					fname, *ushort_ptr));
			break;
		case MXFT_INT:
			int_ptr = (int *) array_ptr;

			*int_ptr = (int) PyInt_AsLong( current_object );

			MX_DEBUG( 2,("%s: INT VALUE = %d",
					fname, *int_ptr));
			break;
		case MXFT_UINT:
			uint_ptr = (unsigned int *) array_ptr;

			*uint_ptr = (unsigned int)
					PyInt_AsLong( current_object );

			MX_DEBUG( 2,("%s: UINT VALUE = %u",
					fname, *uint_ptr));
			break;
		case MXFT_LONG:
			long_ptr = (long *) array_ptr;

			*long_ptr = PyInt_AsLong( current_object );

			MX_DEBUG( 2,("%s: LONG VALUE = %ld", fname, *long_ptr));
			break;
		case MXFT_ULONG:
			ulong_ptr = (unsigned long *) array_ptr;

			*ulong_ptr = (unsigned long)
					PyInt_AsLong( current_object );

			MX_DEBUG( 2,("%s: ULONG VALUE = %lu",
					fname, *ulong_ptr));
			break;
		case MXFT_HEX:
			hex_ptr = (unsigned long *) array_ptr;

			*hex_ptr = (unsigned long)
					PyInt_AsLong( current_object );

			MX_DEBUG( 2,("%s: HEX VALUE = %#lx",
					fname, *hex_ptr));
			break;
		case MXFT_FLOAT:
			float_ptr = (float *) array_ptr;

			*float_ptr = (float) PyFloat_AsDouble( current_object );

			MX_DEBUG( 2,("%s: FLOAT VALUE = %g",
					fname, *float_ptr));
			break;
		case MXFT_DOUBLE:
			double_ptr = (double *) array_ptr;

			*double_ptr = PyFloat_AsDouble( current_object );

			MX_DEBUG( 2,("%s: DOUBLE VALUE = %g",
					fname, *double_ptr));
			break;
		case MXFT_RECORD:
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Writing to an MXFT_RECORD field is not allowed." );
			break;
		case MXFT_RECORDTYPE:
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Writing to an MXFT_RECORDTYPE field is not allowed." );
			break;
		case MXFT_INTERFACE:
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Writing to an MXFT_INTERFACE field is not allowed." );
			break;
		default:
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The record field datatype %ld for field '%s' is unrecognized.",
				field->datatype, field->name);
			break;
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

/*---*/

static mx_status_type
create_tuple_from_field_handler( MX_RECORD *record,
				MX_RECORD_FIELD *field,
				void *handler_data_ptr,
				long *array_indices,
				void *array_ptr,
				long dimension_level )
{
	const char fname[] = "create_tuple_from_field_handler()";

	MPPRIVATE_TUPLE_DATA_STRUCT *tuple_data;
	PyObject **object_array;
	PyObject *parent_tuple, *new_tuple, *new_data_object, *component_object;
	int at_bottom_level;
	long i, k, m, num_elements;
	char *string_value;
	char char_value;
	unsigned char uchar_value;
	short short_value;
	unsigned short ushort_value;
	int int_value;
	unsigned int uint_value;
	long long_value;
	unsigned long ulong_value;
	unsigned long hex_value;
	float float_value;
	double double_value;
	MX_RECORD *record_value;
	long recordtype_value;
	MX_INTERFACE *interface_value;

	MX_DEBUG( 2,("++++++ create_tuple_from_field_handler ++++++"));

	MX_DEBUG( 2,("%s: record = '%s', field = '%s', DIMENSION_LEVEL = %ld",
		fname, record->name, field->name, dimension_level ));

	if ( handler_data_ptr == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"handler_data_ptr is a NULL pointer." );
	}
	if ( array_indices == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"array_indices is a NULL pointer." );
	}
	if ( array_ptr == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"array_ptr is a NULL pointer." );
	}

	MX_DEBUG( 2,("%s: handler_data_ptr = %p", fname, handler_data_ptr));

	tuple_data = (MPPRIVATE_TUPLE_DATA_STRUCT *) handler_data_ptr;

	MX_DEBUG( 2,("%s: tuple_data = %p", fname, tuple_data));

	object_array = tuple_data->object_array;

	MX_DEBUG( 2,("%s: object_array = %p", fname, object_array));

	MX_DEBUG( 2,("%s: object_array[0] = %p", fname, object_array[0]));

#if 0
	if ( field->num_dimensions < 2 ) {
		MX_DEBUG( 2,("%s: field->num_dimensions < 2", fname));
	} else {
		MX_DEBUG( 2,
		("%s: object_array[1] = %p", fname, object_array[1]));
	}
#endif

	/*
	 * Guide to array index use:
	 *
	 * m = field->num_dimensions - dimension_level
	 *
	 * object_array[m] = If we create a new tuple now, save a pointer
	 *                  to it here so that the pointer can be used
	 *                  by tuples at the next level down.
	 *
	 * object_array[m-1] = This is the parent tuple of the object that
	 *                    is created by this invocation of the handler
	 *                    unless m == 0 (which means we are at the
	 *                    top level object).
	 *
	 * k = array_indices[m-1] = This tells us what index value to use 
	 *                      when inserting this object into the parent
	 *                      tuple as in
	 *
	 *       PyTuple_SET_ITEM( object_array[m-1], k, new_object )
	 * 
	 * num_elements = field->dimension[m] = Number of elements in the
         *                                      current dimension.
	 *
	 */

	m = field->num_dimensions - dimension_level;

	if ( m > 0 ) {
		parent_tuple = object_array[m-1];

		if ( PyTuple_Check(parent_tuple) == 0 ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
				"The parent_tuple object %p for m = %ld, "
				"dimension_level = %ld is not a tuple.  "
				"This should never happen and is a "
				"program bug.", parent_tuple, m,
				dimension_level );
		}

		k = array_indices[m-1];
	} else
	if ( m == 0 ) {
		parent_tuple = NULL;

		k = -1;
	} else {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"%s was invoked at dimension level %ld for record '%s', "
		"field '%s' which is larger than the total number of "
		"dimensions %ld.  This should _never_ happen and should "
		"be reported as a bug.", fname, dimension_level,
		record->name, field->name, field->num_dimensions );
	}

	MX_DEBUG( 2,("%s: m = %ld, parent_tuple = %p", fname, m, parent_tuple));

	new_tuple = NULL;
	new_data_object = NULL;

	MX_DEBUG( 2,("%s: field->num_dimensions = %ld, m = %ld, k = %ld",
			fname, field->num_dimensions, m, k));

	for ( i = 0; i < field->num_dimensions; i++ ) {
		MX_DEBUG( 2,(
"%s: array_indices[%ld] = %ld, dimension[%ld] = %ld, object_array[%ld] = %p",
			fname, i, array_indices[i], i, field->dimension[i],
			i, object_array[i]));
	}

	/* Figure out whether or not we have reached the bottom of 
	 * the field data structure or not.
	 */

	at_bottom_level = FALSE;

	if ( dimension_level < 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"%s was invoked at dimension level %ld < 0 for record '%s', "
		"field '%s'.  This should _never_ happen and should be "
		"reported as a bug.", fname, dimension_level,
		record->name, field->name );

	} else
	if ( dimension_level > 1 ) {
		at_bottom_level = FALSE;

	} else
	if ( dimension_level == 1 ) {
		if ( field->datatype == MXFT_STRING ) {
			at_bottom_level = TRUE;
		} else {
			at_bottom_level = FALSE;
		}
	} else
	if ( dimension_level == 0 ) {
		if ( field->datatype == MXFT_STRING ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"%s was invoked at dimension level 0 for record '%s' "
			"and string field '%s'.  This should _never_ happen "
			"and should be reported as a bug.", fname,
			record->name, field->name );
		}
		at_bottom_level = TRUE;
	}

	MX_DEBUG( 2,("%s: at_bottom_level = %d", fname, at_bottom_level));

	if ( at_bottom_level == FALSE ) {

		/* Create a new tuple to represent this row. */

		num_elements = field->dimension[m];

		new_tuple = PyTuple_New( (int) num_elements );

		if ( new_tuple == NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to create a %ld element tuple.",
				num_elements );
		}

		MX_DEBUG( 2,("%s: new_tuple = %p ( %ld elements )",
				fname, new_tuple, num_elements));

		MX_DEBUG( 2,("%s: New tuple size = %d",
				fname, PyTuple_Size(new_tuple)));

		/* Save this tuple for use by the row below us (if any). */

		object_array[m] = new_tuple;

		/* Insert the new tuple into the row above us (if any). */

		if ( parent_tuple != NULL ) {
			MX_DEBUG( 2,("%s: Parent tuple OLD size = %d",
				fname, PyTuple_Size(parent_tuple)));
			MX_DEBUG( 2,
		("%s: Inserting new_tuple = %p into parent %p at index %ld.",
				fname, new_tuple, parent_tuple, k));

			PyTuple_SET_ITEM( parent_tuple, k, new_tuple );

			MX_DEBUG( 2,("%s: Parent tuple size = %d",
				fname, PyTuple_Size(parent_tuple)));
		}

	} else {
		/* at_bottom_level == TRUE */

		MX_DEBUG( 2,("%s: field->datatype = %ld, array_ptr = %p",
			fname, field->datatype, array_ptr));

		/* We have reached the bottom level and need to create
		 * a non-tuple object for the raw data element here.
		 */
		switch ( field->datatype ) {
		case MXFT_STRING:
			string_value = (char *) array_ptr;

			MX_DEBUG( 2,("%s: STRING_VALUE = %p",
					fname, string_value));
			MX_DEBUG( 2,("%s: string value = '%s'",
					fname, string_value));

			new_data_object = PyString_FromString( string_value );
			break;
		case MXFT_CHAR:
			char_value = *(char *) array_ptr;

			MX_DEBUG( 2,("%s: char value = '%c' %#x",
					fname, char_value, char_value));

			new_data_object = PyInt_FromLong( (long) char_value );
			break;
		case MXFT_UCHAR:
			uchar_value = *(unsigned char *) array_ptr;
			MX_DEBUG( 2,("%s: uchar value = '%c' %#x",
					fname, uchar_value, uchar_value));

			new_data_object = PyInt_FromLong( (long) uchar_value );
			break;
		case MXFT_SHORT:
			short_value = *(short *) array_ptr;
			MX_DEBUG( 2,("%s: short value = %hd",
					fname, short_value));

			new_data_object = PyInt_FromLong( (long) short_value );
			break;
		case MXFT_USHORT:
			ushort_value = *(unsigned short *) array_ptr;
			MX_DEBUG( 2,("%s: ushort value = %hu",
					fname, ushort_value));

			new_data_object = PyInt_FromLong( (long) ushort_value );
			break;
		case MXFT_INT:
			int_value = *(int *) array_ptr;
			MX_DEBUG( 2,("%s: int value = %d", fname, int_value));

			new_data_object = PyInt_FromLong( (long) int_value );
			break;
		case MXFT_UINT:
			uint_value = *(unsigned int *) array_ptr;
			MX_DEBUG( 2,("%s: uint value = %u", fname, uint_value));

			new_data_object = PyInt_FromLong( (long) uint_value );
			break;
		case MXFT_LONG:
			long_value = *(long *) array_ptr;
			MX_DEBUG( 2,("%s: long value = %ld",
					fname, long_value));

			new_data_object = PyInt_FromLong( long_value );
			break;
		case MXFT_ULONG:
			ulong_value = *(unsigned long *) array_ptr;
			MX_DEBUG( 2,("%s: ulong value = %lu",
					fname, ulong_value));

			new_data_object = PyInt_FromLong( (long) ulong_value );
			break;
		case MXFT_HEX:
			hex_value = *(unsigned long *) array_ptr;
			MX_DEBUG( 2,("%s: long value = %#lx",
					fname, hex_value));

			new_data_object = PyInt_FromLong( (long) hex_value );
			break;
		case MXFT_FLOAT:
			float_value = *(float *) array_ptr;
			MX_DEBUG( 2,("%s: float value = %g",
					fname, float_value));

			new_data_object =
				PyFloat_FromDouble( (double) float_value );
			break;
		case MXFT_DOUBLE:
			double_value = *(double *) array_ptr;
			MX_DEBUG( 2,("%s: DOUBLE VALUE = %g",
					fname, double_value));

			new_data_object = PyFloat_FromDouble( double_value );
			break;
		case MXFT_RECORD:
			record_value = *(MX_RECORD **) array_ptr;
			MX_DEBUG( 2,("%s: record value = '%s'",
					fname, record_value->name));

			new_data_object
				= PyString_FromString( record_value->name );
			break;
		case MXFT_RECORDTYPE:
			recordtype_value = *(long *) array_ptr;
			MX_DEBUG( 2,
			    ("%s: recordtype value = %ld",
					fname, recordtype_value));

			new_data_object = PyInt_FromLong( recordtype_value );
			break;
		case MXFT_INTERFACE:
			interface_value = *(MX_INTERFACE **) array_ptr;
			MX_DEBUG( 2,
				("%s: interface value = '%s', address = '%s'",
				fname, interface_value->record->name,
				interface_value->address_name));

			new_data_object = PyTuple_New( 2 );

			if ( new_data_object != NULL ) {
				component_object = PyString_FromString(
						interface_value->record->name );

				if ( component_object == NULL ) {
					return mx_error(
					MXE_OUT_OF_MEMORY, fname,
		"Could not allocate memory for the name of interface '%s'.",
						interface_value->record->name );
				}
				PyTuple_SET_ITEM( new_data_object, 0,
							component_object );
				
				component_object = PyString_FromString(
						interface_value->address_name );

				if ( component_object == NULL ) {
					return mx_error(
					MXE_OUT_OF_MEMORY, fname,
	"Could not allocate memory for the address_name of interface '%s'.",
						interface_value->record->name );
				}
				PyTuple_SET_ITEM( new_data_object, 1,
							component_object );
			}
			break;
		default:
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The record field datatype %ld for field '%s' is unrecognized.",
				field->datatype, field->name);
			break;
		}

		if ( new_data_object == NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Could not allocate memory for a new data object "
			"for record '%s', field '%s'",
				record->name, field->name );
		}

		MX_DEBUG( 2,("%s: new_data_object = %p, type = %s",
			fname, new_data_object,
			mx_get_field_type_string( field->datatype ) ));

		MX_DEBUG( 2,("%s: object_array = %p, m = %ld",
			fname, object_array, m));

		MX_DEBUG( 2,("%s: (before) object_array[%ld] = %p",
			fname, m, object_array[m]));

		object_array[m] = new_data_object;

		MX_DEBUG( 2,
			("%s: after assignment of new_data_object.", fname ));

		/* Insert the new tuple into the row above us (if any). */

		if ( parent_tuple != NULL ) {
			MX_DEBUG( 2,("%s: Parent tuple OLD size = %d",
				fname, PyTuple_Size(parent_tuple)));
			MX_DEBUG( 2,
	("%s: Inserting new_data_object = %p into parent %p at index %ld.",
				fname, new_data_object, parent_tuple, k));

			PyTuple_SET_ITEM( parent_tuple, k, new_data_object );

			MX_DEBUG( 2,("%s: Parent tuple size = %d",
				fname, PyTuple_Size(parent_tuple)));
		}
	}
	return MX_SUCCESSFUL_RESULT;
}
	
