/*
 * Name:    mp_user_interrupt.c
 *
 * Purpose: Python module interface to generic MX user interrupt functions.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "Python.h"

#include "mx_util.h"
#include "mx_record.h"
#include "mx_handle.h"
#include "mx_key.h"
#include "mp.h"

/*-------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpPrivate_getch_cmd( PyObject *self, PyObject *args )
{
	PyObject *result;
	int character;

	character = mx_getch();

	result = Py_BuildValue( "i", character );

	return result;
}

MX_EXPORT PyObject *
MpPrivate_kbhit_cmd( PyObject *self, PyObject *args )
{
	PyObject *result;
	int return_value;

	return_value = mx_kbhit();

	result = Py_BuildValue( "i", return_value );

	return result;
}

/*-------------------------------------------------------------------------*/

static PyObject *MpPrivate_user_interrupt_function = NULL;

static int MpPrivate_user_interrupt_handler( void );

MX_EXPORT PyObject *
MpPrivate_user_requested_interrupt_cmd( PyObject *self, PyObject *args )
{
	PyObject *result;
	int user_interrupt;

	user_interrupt = mx_user_requested_interrupt();

	result = Py_BuildValue( "i", user_interrupt );

	return result;
}

MX_EXPORT PyObject *
MpPrivate_set_user_interrupt_function_cmd( PyObject *self, PyObject *args )
{
	static const char cname[] = "set_user_interrupt_function";

	PyObject *function_object;
	int parse_status;

	parse_status = PyArg_ParseTuple( args, "O", &function_object );

	if ( parse_status == 0 )
		return NULL;

	/* Get rid of any old function object. */

	if ( MpPrivate_user_interrupt_function != NULL ) {
		Py_DECREF( MpPrivate_user_interrupt_function );

		MpPrivate_user_interrupt_function = NULL;
	}

	/* If the argument passed is None, then change the user interrupt
	 * handler back to the default handler and exit.
	 */

	if ( function_object == Py_None ) {
		mx_set_user_interrupt_function( NULL );

		Py_INCREF(Py_None);
		return Py_None;
	}

	/* Is the argument passed a callable object? */

	if ( PyCallable_Check( function_object ) == 0 ) {
		MpPrivateErr_SetString( PyExc_RuntimeError,
"The argument passed to '%s' is not a function or other callable object.",
			cname );

		return NULL;
	}

	/* Save the supplied function argument. */

	MpPrivate_user_interrupt_function = function_object;

	Py_INCREF( MpPrivate_user_interrupt_function );

	/* Install an MX user interrupt handler. */

	mx_set_user_interrupt_function( MpPrivate_user_interrupt_handler );

	Py_INCREF(Py_None);

	return Py_None;
}

static int
MpPrivate_user_interrupt_handler( void )
{
	static const char fname[] = "MpPrivate_user_interrupt_handler()";

	PyObject *arglist;
	PyObject *result;
	int parse_status, user_interrupt;

	if ( MpPrivate_user_interrupt_function == NULL ) {
		fprintf( stderr, "%s: Attempted to invoke the user interrupt "
		"handler but the MpPrivate_user_interrupt_function pointer "
		"was NULL.", fname );

		return (-1);
	}

	arglist = Py_BuildValue("()");

	if ( arglist == NULL ) {
		fprintf( stderr, "%s: arglist was NULL.\n", fname );

		PyErr_Clear();

		return (-1);
	}

	result = PyEval_CallObject( MpPrivate_user_interrupt_function,
					arglist );

	Py_DECREF( arglist );

	if ( result == NULL ) {

#if ( PY_MAJOR_VERSION >= 2 )
		PyErr_WriteUnraisable( MpPrivate_user_interrupt_function );
#endif
		PyErr_Clear();

		return (-1);
	}

	parse_status = PyArg_ParseTuple( result, "i", &user_interrupt );

	if ( parse_status == 0 ) {

#if ( PY_MAJOR_VERSION >= 2 )
		PyErr_WriteUnraisable( MpPrivate_user_interrupt_function );
#endif
		PyErr_Clear();

		return (-1);
	}

	Py_DECREF( result );

	return user_interrupt;
}

